/* -*-  Mode:C; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1999, 2000 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CMAUTHORS which should have accompanied this software 
 * for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CMAUTHORS file.
 */
/*
 * An implementation of the CM API for Linux, conforming roughly to
 * draft-ietf-ecm-cm-00.txt of the ECM working group.  This draft is 
 * available from http://nms.lcs.mit.edu/papers/draft-ietf-ecm-cm-00.txt
 * Applications should link with this library and also include cmapp.h
 */
#include <sys/socket.h>
#include <stdio.h>
#include <unistd.h>
#include <netinet/in.h>
#include "cmapp.h"

void (*_cmapp_send_handler)(int) = NULL;
void (*_cmapp_update_handler)(int, cmquery) = NULL;

/*
 * Open a CM flow to the destination sockaddr_in.  This is the stream_info 
 * described in Section 4.1 of the IETF draft.
 * cm_open() does a socket() and connect() of a control socket that the 
 * CM library uses to communicate with the kernel.
 * The srcaddr sockaddr is needed to handle multi-homed hosts correctly.
 */
int cm_open(struct sockaddr_in *srcaddr, struct sockaddr_in *dstaddr)
{
	int id = -1, err;
	struct sockaddr_in addr;
	/*
	 * This is the control socket to communicate with the CM in the kernel.
	 * We do not use signals for callbacks for a variety of reasons.
	 */
	id = socket(AF_INET, SOCK_CM, IPPROTO_CM);
	if (id == -1) {
#ifdef DEBUG
		perror("cm_open: control socket's socket call");
#endif
		return -1;
	}
	if (dstaddr) {
		err = connect(id, dstaddr, sizeof(addr));
		if (err == -1) {
#ifdef DEBUG
			perror("cm_open: control socket's connect");
#endif
			return -1;
		}
	}
	return id;
}

/*
 * Close and clean up flow state.  Section 4.1.
 * XXX incomplete: need ioctl for clean up.
 */
int cm_close(int id)
{
	return close(id);
}

/*
 * Get (path) MTU.  This decides how much data can be transmitted in each
 * transmission when a callback (cmapp_send) is done.  Section 4.1.
 */
int cm_mtu(int cmid)
{
	int mtu = -1, err;

	err = ioctl(cmid, SIOCGIFMTU, &mtu);
	if (err == -1) {
#ifdef DEBUG
		perror("cm_mtu: ioctl failed");
#endif
		return -1;
	}
	return mtu;
}

/*
 * Ask to send up to MTU bytes.  Section 4.2 (#1).
 */
int cm_request(int cmid)
{
	int err;
 
	err = ioctl(cmid, SIOCCMREQUEST, NULL);
#ifdef DEBUG
	if (err == -1) {
		perror("cm_request: ioctl failed");
	}
#endif
	return err;
}

/*
 * Undocumented bulk request API.
 */
int cm_bulk_request(int cmid, int *list)
{
	int err;
 
	err = ioctl(cmid, SIOCCMBULKREQUEST, list);
#ifdef DEBUG
	if (err == -1) {
		perror("cm_bulk_request: ioctl failed");
	}
#endif
	return err;
}

/*
 * Ask what the path statistics are.  Section 4.4 (used in 4.2 #2).
 */
int cm_query(int id, int *ready, cmquery *cmquery)
{
	int err;

	if (ready) {
		err = ioctl(id, SIOCCMREADY, ready);
		if (err == -1) {
#ifdef DEBUG
			perror("cm_query: ready ioctl failed");
#endif
			return -1;
		}
	}
	if (cmquery) {
		err = ioctl(id, SIOCCMQUERY, cmquery);
		if (err == -1) {
#ifdef DEBUG
			perror("cmtest: cm_query: cmquery ioctl");
#endif
			return -1;
		}
	}
	return err;
}

/*
 * Ask what the path statistics are for ALL the open streams on this process.
 * cmfdquery should be an array of cmquery's that is big enough to hold
 * cmquerys for *all* open cm sockets.
 * The return value is the number of sockets reported on.
 * This API is undocumented in the draft.
 */
int cm_bulk_query(int id, cmfdquery *cmfdquery)
{
	int err;

	err = ioctl(id, SIOCCMBULKQUERY, cmfdquery);
#ifdef DEBUG
	if (err == -1) {
		perror("cmtest: cm_query: cmfdquery ioctl");
	}
#endif
	return err;
}

/*
 * Set callback thresholds so the CM knows when to call cmapp_update().
 */
int cm_setthresh(int id, cmthresh *cmthresh)
{
	int err;
	err = ioctl(id, SIOCCMSETTHRESHM, cmthresh);
	return err;
}

/* 
 * For conformance with IETF draft.
 */
inline int cm_thresh(int id, cmthresh *cmthresh)
{
	return cm_setthresh(id, cmthresh);
}

/*
 * Get current threshold settings.
 */
int cm_getthresh(int id, cmthresh *cmthresh)
{
	int err;
	err = ioctl(id, SIOCCMGETTHRESHM, cmthresh);
#ifdef DEBUG
	if (err == -1) {
		perror("cmtest: cm_getthresh: ioctl");
	}
#endif
	return err;
}

/*
 * Tell the CM what we know about recent transmissions.  Section 4.3.
 * nrecd and nloss are in bytes; rtt is in microseconds.
 */
int cm_update(int id, int nrecd, int nloss, int lossmode,
			unsigned int rtt)
{
	int err;
	struct cmupdate cmupdate;

	cmupdate.nrecd = nrecd;
	cmupdate.nloss = nloss;
	cmupdate.lossmode = lossmode;
	cmupdate.rtt = rtt;
	err = ioctl(id, SIOCCMUPDATE, &cmupdate);
#ifdef DEBUG
	if (err == -1) {
		perror("ioctl in cm_update");
	}
#endif
	return err;
}

/*
 * Bulk cm_update() call for efficiency.
 * Tell the CM what we know about recent transmissions on all streams.
 * The format of list is:
 *	number of sockets described
 *      fd for 1st socket
 *	number of bytes received
 *	number of bytes lost
 *	lossmode
 *	rtt
 *	...
 * Undocumented in draft.
 */
int cm_bulk_update(int id, int *list)
{
	int err;
	err = ioctl(id, SIOCCMBULKUPDATE, list);
#ifdef DEBUG
	if (err == -1) {
		perror("cmtest: cm_update: ioctl");
	}
#endif
	return err;
}

/*
 * This is here simply for illustration.  It is NOT to be called unless
 * there are some extenuating circumstances.  cm_notify() SHOULD be
 * called from IP output code.
 */
int cm_notify(int id, int nsent)
{
	int err;

	err = ioctl(id, SIOCCMNOTIFY, &nsent);
#ifdef DEBUG
	if (err == -1) {
		perror("cmtest: cm_notify: ioctl");
	}
#endif
	return err;
}

/*
 * Tell CM about data that has been sent.
 * The format of list is:
 *	number of sockets described
 *      fd for 1st socket
 *	number of bytes sent for 1st socket
 *	...
 * Undocumented in draft.
 */
int cm_bulk_notify(int id, int *list)
{
	int err;

	err = ioctl(id, SIOCCMBULKNOTIFY, list);
#ifdef DEBUG
	if (err == -1) {
		perror("cmtest: cm_notify: ioctl");
	}
#endif
	return err;
}

/*
 * Macroflow API functions.  Section 4.5.
 */
int cm_getmacroflow(int id)
{
	int err;
	int mflow = -3;
	err = ioctl(id, SIOCCMGETMFLOW, &mflow);
	if (err == -1) {
#ifdef DEBUG
		perror("cmtest: cm_getmflow: ioctl");
#endif
		return -1;
	}
	return mflow;
}

int cm_setmacroflow(int id, int mflow)
{
	int err;
	err = ioctl(id, SIOCCMSETMFLOW, &mflow);
	return err;
}

int cm_setassoc(int id, int *list)
{
	int err;
	err = ioctl(id, SIOCCMSETASSOC, list);
	return err;
}

int cm_getassoc(int id, int *list)
{
	int err;
	err = ioctl(id, SIOCCMGETASSOC, list);
	return err;
}

int cm_setcongalg(int id, int type)
{
	int err;
	err = ioctl(id, SIOCCMSETCONGALG, type);
	return err;
}

void cm_register_send(void (*handler)(int))
{
	_cmapp_send_handler = handler;
}

void cm_register_update(void (*handler)(int, cmquery))
{
	_cmapp_update_handler = handler;
}

/*
 * In the future, this function will use the batch interface
 * to dispatch multiple updates and sends from a single query.
 */

void cm_dispatch(int fd, int select_read, int select_write, int select_except)
{
	if ((select_write && _cmapp_send_handler != NULL)) {
		_cmapp_send_handler(fd);
	}

	if ((select_except && _cmapp_update_handler != NULL)) {
		cmquery cminfo;
		if (cm_query(fd, NULL, &cminfo) == 0) {
			_cmapp_update_handler(fd, cminfo);
		}
	}
}
